@@ -1,18 +1,22 @@
+## no critic (Moose::RequireCleanNamespace, Moose::RequireMakeImmutable)
 package MooseX::Params::Validate;
-# git description: v0.18-5-g3bcb95b
 
-$MooseX::Params::Validate::VERSION = '0.19';
 use strict;
 use warnings;
 
+our $VERSION = '0.21';
+
 use Carp 'confess';
 use Devel::Caller 'caller_cv';
 use Scalar::Util 'blessed', 'refaddr', 'reftype';
 
-use Moose 0.58 ();
+use Moose 2.1200 ();
+use Moose::Util qw( throw_exception );
 use Moose::Util::TypeConstraints
     qw( find_type_constraint class_type role_type );
-use Params::Validate 0.88 ();
+use MooseX::Params::Validate::Exception::ValidationFailedForTypeConstraint;
+use Params::Validate 1.15 ();
+
 use Sub::Exporter -setup => {
     exports => [
         qw( validated_hash validated_list pos_validated_list validate validatep )
@@ -36,14 +40,16 @@ sub validated_hash {
         ( ref $CACHED_SPECS{$cache_key} eq 'HASH' )
             || confess
             "I was expecting a HASH-ref in the cached $cache_key parameter"
-            . " spec, you are doing something funky, stop it!";
+            . ' spec, you are doing something funky, stop it!';
         %spec = %{ $CACHED_SPECS{$cache_key} };
     }
     else {
         my $should_cache = delete $spec{MX_PARAMS_VALIDATE_NO_CACHE} ? 0 : 1;
 
-        $spec{$_} = _convert_to_param_validate_spec( $spec{$_} )
-            foreach keys %spec;
+        $spec{$_} = _convert_to_param_validate_spec(
+            $spec{$_},
+            qq{The '$_' parameter},
+        ) foreach keys %spec;
 
         $CACHED_SPECS{$cache_key} = \%spec
             if $should_cache;
@@ -86,7 +92,7 @@ sub validated_list {
         ( ref $CACHED_SPECS{$cache_key} eq 'ARRAY' )
             || confess
             "I was expecting a ARRAY-ref in the cached $cache_key parameter"
-            . " spec, you are doing something funky, stop it!";
+            . ' spec, you are doing something funky, stop it!';
         %spec         = %{ $CACHED_SPECS{$cache_key}->[0] };
         @ordered_spec = @{ $CACHED_SPECS{$cache_key}->[1] };
     }
@@ -95,8 +101,10 @@ sub validated_list {
 
         @ordered_spec = grep { exists $spec{$_} } @spec;
 
-        $spec{$_} = _convert_to_param_validate_spec( $spec{$_} )
-            foreach keys %spec;
+        $spec{$_} = _convert_to_param_validate_spec(
+            $spec{$_},
+            qq{The '$_' parameter},
+        ) foreach keys %spec;
 
         $CACHED_SPECS{$cache_key} = [ \%spec, \@ordered_spec ]
             if $should_cache;
@@ -145,14 +153,18 @@ sub pos_validated_list {
         ( ref $CACHED_SPECS{$cache_key} eq 'ARRAY' )
             || confess
             "I was expecting an ARRAY-ref in the cached $cache_key parameter"
-            . " spec, you are doing something funky, stop it!";
+            . ' spec, you are doing something funky, stop it!';
         @pv_spec = @{ $CACHED_SPECS{$cache_key} };
     }
     else {
         my $should_cache = exists $extra{MX_PARAMS_VALIDATE_NO_CACHE} ? 0 : 1;
 
-        # prepare the parameters ...
-        @pv_spec = map { _convert_to_param_validate_spec($_) } @spec;
+        @pv_spec = map {
+            _convert_to_param_validate_spec(
+                $spec[$_],
+                'Parameter #' . ( $_ + 1 )
+                )
+        } 0 .. $#spec;
 
         $CACHED_SPECS{$cache_key} = \@pv_spec
             if $should_cache;
@@ -185,7 +197,9 @@ sub _cache_key {
 }
 
 sub _convert_to_param_validate_spec {
-    my ($spec) = @_;
+    my $spec = shift;
+    my $id   = shift;
+
     my %pv_spec;
 
     $pv_spec{optional} = $spec->{optional}
@@ -202,15 +216,14 @@ sub _convert_to_param_validate_spec {
 
     my $constraint;
     if ( defined $spec->{isa} ) {
-        $constraint
-             = _is_tc( $spec->{isa} )
+        $constraint = _is_tc( $spec->{isa} )
             || Moose::Util::TypeConstraints::find_or_parse_type_constraint(
             $spec->{isa} )
             || class_type( $spec->{isa} );
     }
     elsif ( defined $spec->{does} ) {
         $constraint
-            = _is_tc( $spec->{isa} )
+            = _is_tc( $spec->{does} )
             || find_type_constraint( $spec->{does} )
             || role_type( $spec->{does} );
     }
@@ -221,9 +234,19 @@ sub _convert_to_param_validate_spec {
     if ($constraint) {
         $pv_spec{constraint} = $constraint;
 
+        my $cb = sub {
+            return 1 if $constraint->check( $_[0] );
+            die
+                MooseX::Params::Validate::Exception::ValidationFailedForTypeConstraint
+                ->new(
+                parameter => $id,
+                type      => $constraint,
+                value     => $_[0],
+                );
+        };
+
         $pv_spec{callbacks}
-            { 'checking type constraint for ' . $constraint->name }
-            = sub { $constraint->check( $_[0] ) };
+            { 'checking type constraint for ' . $constraint->name } = $cb;
     }
 
     delete $pv_spec{coerce}
@@ -237,8 +260,8 @@ sub _is_tc {
 
     return $maybe_tc
         if defined $maybe_tc
-            && blessed $maybe_tc
-            && $maybe_tc->isa('Moose::Meta::TypeConstraint');
+        && blessed $maybe_tc
+        && $maybe_tc->isa('Moose::Meta::TypeConstraint');
 }
 
 sub _caller_name {
@@ -247,6 +270,8 @@ sub _caller_name {
     return ( caller( 2 + $depth ) )[3];
 }
 
+no Moose::Util::TypeConstraints;
+
 1;
 
 # ABSTRACT: an extension of Params::Validate using Moose's types
@@ -261,7 +286,7 @@ MooseX::Params::Validate - an extension of Params::Validate using Moose's types
 
 =head1 VERSION
 
-version 0.19
+version 0.21
 
 =head1 SYNOPSIS
 
@@ -431,6 +456,17 @@ below, simply pass them after the list of parameter validation specs:
 
 =back
 
+=head1 EXCEPTION FOR FAILED VALIDATION
+
+If a type constraint check for a parameter fails, then the error is thrown as
+a L<MooseX::Params::Validate::Exception::ValidationFailedForTypeConstraint>
+object. When stringified, this object will use the error message generated by
+the type constraint that failed.
+
+Other errors are simply percolated up from L<Params::Validate> as-is, and are
+not turned into exception objects. This may change in the future (or more
+likely, L<Params::Validate> may start throwing objects of its own).
+
 =head1 ALLOWING EXTRA PARAMETERS
 
 By default, any parameters not mentioned in the parameter spec cause this
@@ -511,7 +547,7 @@ email at bug-moosex-params-validate@rt.cpan.org.
 
 =item *
 
-Stevan Little <stevan.little@iinteractive.com>
+Stevan Little <stevan@cpan.org>
 
 =item *
 
@@ -541,7 +577,7 @@ Karen Etheridge <ether@cpan.org>
 
 =head1 COPYRIGHT AND LICENSE
 
-This software is copyright (c) 2013 - 2014 by Stevan Little <stevan.little@iinteractive.com>.
+This software is copyright (c) 2013 - 2015 by Stevan Little <stevan@cpan.org>.
 
 This is free software; you can redistribute it and/or modify it under
 the same terms as the Perl 5 programming language system itself.
